/**
    @filename   :   epd7in5b.cpp
    @brief      :   Implements for e-paper library
    @author     :   Yehui from Waveshare

    Copyright (C) Waveshare     August 10 2017

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documnetation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to  whom the Software is
   furished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS OR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
   THE SOFTWARE.

   Add profile
*/

#include <stdlib.h>
#include "epd7in5b.h"

Epd::~Epd() {
};

Epd::Epd() {
  reset_pin = RST_PIN;
  dc_pin = DC_PIN;
  cs_pin = CS_PIN;
  busy_pin = BUSY_PIN;
  width = EPD_WIDTH;
  height = EPD_HEIGHT;
};

// test color profile for all all eight pixel values
//
//                                      cnt    1-2-3-4    5-6-7-8      f1    f2    f3    f4    f5    f6    f7    f8
//                                                                     wht         blk         wht         blk
//                                       1     2     3                  4     5     6     7     8     9     10    11
const unsigned char lut_vcom_buff[]  = {0x04, B01001000, B01001001,   0x3f, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x5f, 0x3f,
                                        0x01, B00000000, B00000000,   0x2f, 0x2f, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00
                                       };
//

//                                       cnt   1-2    3-4   5-6   7-8   f1    f2    f3    f4    f5    f6    f7    f8
//                                        1     2     3     4     5     6     7     8     9     10    11    12    13
const unsigned char lut_White_buff[] = {0x04, 0x00, 0x00, 0x00, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x5f, 0x3f,
                                        0x01, 0x20, 0x20, 0x00, 0x00, 0x2f, 0x2f, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00,
                                       };

//                                       cnt   1-2    3-4   5-6   7-8   f1    f2    f3    f4    f5    f6    f7    f8
//                                        1     2     3     4     5     6     7     8     9     10    11    12    13
const unsigned char lut_Black_buff[] = {0x04, 0x00, 0x00, 0x00, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x5f, 0x3f,
                                        0x01, 0x21, 0x20, 0x00, 0x00, 0x2f, 0x2f, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00,
                                        0x05, 0x43, 0x00, 0x00, 0x00, 0x1f, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                                        0x01, 0x11, 0x00, 0x00, 0x00, 0x0f, 0x2f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                                       };

const unsigned char lut_Gray1_buff[] = {0x04, 0x00, 0x00, 0x00, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x5f, 0xf0,
                                        0x01, 0x12, 0x10, 0x10, 0x0f, 0x0f, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                                       };

const unsigned char lut_Gray2_buff[] = {0x00, 0x00, 0x40, 0x00, 0x40, 0x3f, 0x0f, 0x3f, 0x0f, 0x3f, 0x0f, 0x0f, 0x0f
                                       }; 

const unsigned char lut_Red0_buff[]  = {0x04, 0x00, 0x00, 0x00, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x53f, 0x3f,
                                        0x01, 0x21, 0x20, 0x00, 0x00, 0x2f, 0x2f, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00,
                                        0x05, 0x34, 0x00, 0x00, 0x00, 0x1f, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                                        0x01, 0x30, 0x00, 0x00, 0x00, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                                       };

const unsigned char lut_Red1_buff[]  = {0x04, 0x00, 0x00, 0x00, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x5f, 0x3f,
                                        0x01, 0x21, 0x20, 0x00, 0x00, 0x2f, 0x2f, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00,
                                        0x05, 0x34, 0x00, 0x00, 0x00, 0x1f, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                                        0x01, 0x30, 0x00, 0x00, 0x00, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                                       };
                                       
const unsigned char lut_Red2_buff[]  = {0x04, 0x00, 0x00, 0x00, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x5f, 0x3f,
                                        0x01, 0x21, 0x20, 0x00, 0x00, 0x2f, 0x2f, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00,
                                        0x05, 0x34, 0x00, 0x00, 0x00, 0x1f, 0x06, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                                        0x01, 0x30, 0x00, 0x00, 0x00, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                                       };
                                       
const unsigned char lut_Red3_buff[]  = {0x04, 0x00, 0x00, 0x00, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x5f, 0x3f,
                                        0x01, 0x21, 0x20, 0x00, 0x00, 0x2f, 0x2f, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00,
                                        0x05, 0x34, 0x00, 0x00, 0x00, 0x1f, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                                        0x01, 0x30, 0x00, 0x00, 0x00, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                                       };

// Short black and white profile only
//                                      cnt    1-2-3-4    5-6-7-8      f1    f2    f3    f4    f5    f6    f7    f8
//                                                                     wht         blk         wht         blk
//                                       1     2     3                  4     5     6     7     8     9     10    11
const unsigned char lutBW_vcom_buff[22]  = {0x02, B01001000, B01001000,   0x3f, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x5f, 0x3f,
                                        0x01, B00000000, B00000000,   0x2f, 0x2f, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00};
//

//                                       cnt   1-2    3-4   5-6   7-8   f1    f2    f3    f4    f5    f6    f7    f8
//                                        1     2     3     4     5     6     7     8     9     10    11    12    13
const unsigned char lutBW_White_buff[26] = {0x02, 0x00, 0x00, 0x00, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x5f, 0x3f,
                                        0x01, 0x20, 0x20, 0x00, 0x00, 0x2f, 0x2f, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00,
                                        };

//                                       cnt   1-2    3-4   5-6   7-8   f1    f2    f3    f4    f5    f6    f7    f8
//                                        1     2     3     4     5     6     7     8     9     10    11    12    13
const unsigned char lutBW_Black_buff[52] = {0x02, 0x00, 0x00, 0x00, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x3f, 0x00, 0x5f, 0x3f,
                                        0x01, 0x11, 0x10, 0x00, 0x00, 0x2f, 0x2f, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00,
                                        };

const unsigned char lutBW_Gray1_buff[26] =  {0x00,};
const unsigned char lutBW_Gray2_buff[13] = {0x00,};
const unsigned char lutBW_Red0_buff[52]  = {0x00,};
const unsigned char lutBW_Red1_buff[52]  = {0x00,};
const unsigned char lutBW_Red2_buff[52]  = {0x00,};
const unsigned char lutBW_Red3_buff[52]  = {0x00,};




void Epd::FillLutCom(unsigned char lutid, const unsigned char* pbuff, int length, unsigned char fill)
{
  int count;
  SendCommand(lutid);
  for (count = 0; count < length; count++) {
    SendData(pbuff[count]);
  }
  for (; count < (20 * 11); count++) {
    SendData(fill);
  }
}

void Epd::FillLut(unsigned char lutid, const unsigned char* pbuff, int length, unsigned char fill)
{
  int count;
  SendCommand(lutid);
  for (count = 0; count < length; count++) {
    SendData(pbuff[count]);
  }
  for (; count < (20 * 13); count++) {
    SendData(fill);
  }
}



int Epd::Init(void) {

  //Dummy init
  if (IfInit() != 0) {
    return -1;
  }
  Reset();

  delay(1);
  SendCommand(BOOSTER_SOFT_START);//x06
  SendData(0xd7); // c7
  SendData(0xd7); // cc
  SendData(0x2f); // 28

  delay(1);
  SendCommand(POWER_SETTING); //x01
  SendData(0x37);
  SendData(0x00);
  SendData(0x09);
  SendData(0x3f);

  delay(1);
  SendCommand(POWER_ON); //x04
  Serial.print ("power up ");
  WaitUntilIdle();

  SendCommand(PANEL_SETTING);//x00
  SendData(0xcF); //(c7 upsidedown)
  SendData(0x00);  //Run with oiut then with

  SendCommand(VCOM_AND_DATA_INTERVAL_SETTING);//x50
  SendData(0x77);

  SendCommand(PLL_CONTROL);//x30
  SendData(0x3a);

  SendCommand(TCON_RESOLUTION);//x61
  SendData(0x02);     //source 640
  SendData(0x80);
  SendData(0x01);     //gate 384
  SendData(0x80);

  SendCommand(VCM_DC_SETTING);//x82
  SendData(0x1E);      //decide by LUT file

  SendCommand(DISPLAY_REFRESH);
  DelayMs(300);
  //Reset();

  WaitUntilIdle();

  // ----------------------------------------------
  // Second initialization

  Reset();

  delay(1);
  SendCommand(BOOSTER_SOFT_START);//x06
  SendData(0xd7);
  SendData(0xd7);
  SendData(0x2f);

  delay(1);
  SendCommand(POWER_SETTING); //x01
  SendData(0x37);
  SendData(0x00);
  SendData(0x09);
  SendData(0x3f);


  delay(1);
  SendCommand(POWER_ON); //x04
  Serial.print ("power up ");
  WaitUntilIdle();

  SendCommand(PANEL_SETTING);//x00
  SendData(0xcF); //(c7 upsidedown)
  SendData(0x80);  //Run with oiut then with

  SendCommand(VCOM_AND_DATA_INTERVAL_SETTING);//x50
  SendData(0x77);

  SendCommand(PLL_CONTROL);//x30
  SendData(0x3a);

  SendCommand(TCON_RESOLUTION);//x61
  SendData(0x02);     //source 640
  SendData(0x80);
  SendData(0x01);     //gate 384
  SendData(0x80);

  SendCommand(VCM_DC_SETTING);//x82
  SendData(0x1E);      //decide by LUT file

  //------------ Load color profile
  unsigned char fill = 0x00;
  FillLutCom(LUT_FOR_VCOM, lut_vcom_buff, sizeof(lut_vcom_buff) / sizeof(lut_vcom_buff[0]), fill);
  FillLut   (LUT_BLACK   , lut_Black_buff, sizeof(lut_Black_buff) / sizeof(lut_Black_buff[0]), fill);
  FillLut   (LUT_WHITE   , lut_White_buff, sizeof(lut_White_buff) / sizeof(lut_White_buff[0]), fill);
  FillLut   (LUT_GRAY_1  , lut_Gray1_buff, sizeof(lut_Gray1_buff) / sizeof(lut_Gray1_buff[0]), fill);
  FillLut   (LUT_GRAY_2  , lut_Gray2_buff, sizeof(lut_Gray2_buff) / sizeof(lut_Gray2_buff[0]), fill);
  FillLut   (LUT_RED_0   , lut_Red0_buff, sizeof(lut_Red0_buff) / sizeof(lut_Red0_buff[0]), fill);
  FillLut   (LUT_RED_1   , lut_Red1_buff, sizeof(lut_Red1_buff) / sizeof(lut_Red1_buff[0]), fill);
  FillLut   (LUT_RED_2   , lut_Red2_buff, sizeof(lut_Red2_buff) / sizeof(lut_Red2_buff[0]), fill);
  FillLut   (LUT_RED_3   , lut_Red3_buff, sizeof(lut_Red3_buff) / sizeof(lut_Red3_buff[0]), fill);

// Shorter Black and White prifile 
//  FillLutCom(LUT_FOR_VCOM, lutBW_vcom_buff, sizeof(lutBW_vcom_buff) / sizeof(lutBW_vcom_buff[0]), fill);
//  FillLut   (LUT_BLACK   , lutBW_Black_buff, sizeof(lutBW_Black_buff) / sizeof(lutBW_Black_buff[0]), fill);
//  FillLut   (LUT_WHITE   , lutBW_White_buff, sizeof(lutBW_White_buff) / sizeof(lutBW_White_buff[0]), fill);
//  FillLut   (LUT_GRAY_1  , lutBW_Gray1_buff, sizeof(lutBW_Gray1_buff) / sizeof(lutBW_Gray1_buff[0]), fill);
//  FillLut   (LUT_GRAY_2  , lutBW_Gray2_buff, sizeof(lutBW_Gray2_buff) / sizeof(lutBW_Gray2_buff[0]), fill);
//  FillLut   (LUT_RED_0   , lutBW_Red0_buff, sizeof(lutBW_Red0_buff) / sizeof(lutBW_Red0_buff[0]), fill);
//  FillLut   (LUT_RED_1   , lutBW_Red1_buff, sizeof(lutBW_Red1_buff) / sizeof(lutBW_Red1_buff[0]), fill);
//  FillLut   (LUT_RED_2   , lutBW_Red2_buff, sizeof(lutBW_Red2_buff) / sizeof(lutBW_Red2_buff[0]), fill);
//  FillLut   (LUT_RED_3   , lutBW_Red3_buff, sizeof(lutBW_Red3_buff) / sizeof(lutBW_Red3_buff[0]), fill);

  return 0;
}

/**
    @brief: basic function for sending commands
*/
void Epd::SendCommand(unsigned char command) {
  DigitalWrite(dc_pin, LOW);
  SpiTransfer(command);
}

/**
    @brief: basic function for sending data
*/
void Epd::SendData(unsigned char data) {
  DigitalWrite(dc_pin, HIGH);
  SpiTransfer(data);
}

/**
    @brief: Wait until the busy_pin goes HIGH
*/
void Epd::WaitUntilIdle(void) {
  unsigned long start = micros();
  while (1)
  {

    delay(1);
    if (DigitalRead(busy_pin) != 0) break;
    if (micros() - start > 60000000)
    {
      Serial.println("Busy Timeout!");
      break;
    }
  }

  unsigned long elapsed = micros() - start;
  Serial.print("Wait: ");
  Serial.println(elapsed);
}

/**
    @brief: module reset.
            often used to awaken the module in deep sleep,
            see Epd::Sleep();
*/
void Epd::Reset(void) {
  Serial.println("reset");
  DigitalWrite(reset_pin, LOW);                //module reset
  DelayMs(200);
  DigitalWrite(reset_pin, HIGH);
  DelayMs(200);
}

void Epd::DisplayFrame(const unsigned char** image_data) {
  unsigned char temp1, temp2;

  //SendCommand(0x91);
  SendCommand(DATA_START_TRANSMISSION_1);
  /**
      Size of a single array cannot be larger than 32K in AVR GCC, therefore
      you have to split the image data (61440 bytes in total) into 2 parts
  */
  for (int image_data_part = 0; image_data_part < 2; image_data_part++) {
    for (long i = 0; i < 30720; i++) {
      temp1 = pgm_read_byte(image_data[image_data_part] + i);
      for (unsigned char j = 0; j < 4; j++) {
        if ((temp1 & 0xC0) == 0xC0) {
          temp2 = 0x03;                       // white
        } else if ((temp1 & 0xC0) == 0x00) {
          temp2 = 0x00;                       // black
        } else {
          temp2 = 0x04;                       // red
        }
        temp2 <<= 4;
        temp1 <<= 2;
        j++;
        if ((temp1 & 0xC0) == 0xC0) {
          temp2 |= 0x03;                      // white
        } else if ((temp1 & 0xC0) == 0x00) {
          temp2 |= 0x00;                      // black
        } else {
          temp2 |= 0x04;                      // red
        }
        temp1 <<= 2;
        SendData(temp2);
      }
    }
  }
  SendCommand(DISPLAY_REFRESH);
  DelayMs(300);
  //Reset();

  WaitUntilIdle();
}

void Epd::DisplayBWFrame(const unsigned char** image_data) {
  unsigned char temp1, temp2;

  //SendCommand(0x91);
  SendCommand(DATA_START_TRANSMISSION_1);
  /**
      Size of a single array cannot be larger than 32K in AVR GCC, therefore
      you have to split the image data (61440 bytes in total) into 2 parts
  */
  for (long i = 0; i < 30720; i++) {
    temp1 = pgm_read_byte(image_data[1] + i);
    temp2 = 0; // all black
    if ((temp1 & 0x80) == 0x80)  temp2 |= 0x30;
    if ((temp1 & 0x40) == 0x40)  temp2 |= 0x03;
    SendData(temp2);
    temp2 = 0; // all black
    if ((temp1 & 0x20) == 0x20)  temp2 |= 0x30;
    if ((temp1 & 0x10) == 0x10)  temp2 |= 0x03;
    SendData(temp2);
    temp2 = 0; // all black
    if ((temp1 & 0x08) == 0x08)  temp2 |= 0x30;
    if ((temp1 & 0x04) == 0x04)  temp2 |= 0x03;
    SendData(temp2);
    temp2 = 0; // all black
    if ((temp1 & 0x02) == 0x02)  temp2 |= 0x30;
    if ((temp1 & 0x01) == 0x01)  temp2 |= 0x03;
    SendData(temp2);
  }
  DelayMs(5);
  SendCommand(DISPLAY_REFRESH);
  DelayMs(1);
  //Reset();

  WaitUntilIdle();
}

void Epd::DisplayPattern(void) {

  unsigned char pattern = 0;
  SendCommand(DATA_START_TRANSMISSION_1);
  for (long i = 0; i < 122880; i++) {
    pattern =  (i >> 3) & 0x07;
    pattern |= (pattern << 4);

    SendData(pattern);
  }
  DelayMs(3);
  SendCommand(DISPLAY_REFRESH);
  DelayMs(1);
  WaitUntilIdle();
}

void Epd::Clean(void) {
  SendCommand(DATA_START_TRANSMISSION_1);
  for (long i = 0; i < 122880; i++) {
    SendData(0x33);
  }
  SendCommand(DISPLAY_REFRESH);
  DelayMs(100);
  WaitUntilIdle();
}

void Epd::DisplayOneQuarterFrame(const unsigned char* image_data) {
  unsigned char temp1, temp2;
  SendCommand(DATA_START_TRANSMISSION_1);


  for (long i = 0; i < 192; i++) {
    for (long k = 0; k < 80; k++) {
      temp1 = pgm_read_byte(image_data + i * 80 + k);
      for (unsigned char j = 0; j < 4; j++) {
        if ((temp1 & 0xC0) == 0xC0) {
          temp2 = 0x03;                       // white
        } else if ((temp1 & 0xC0) == 0x00) {
          temp2 = 0x00;                       // black
        } else {
          temp2 = 0x04;                       // red
        }
        temp2 <<= 4;
        temp1 <<= 2;
        j++;
        if ((temp1 & 0xC0) == 0xC0) {
          temp2 |= 0x03;                      // white
        } else if ((temp1 & 0xC0) == 0x00) {
          temp2 |= 0x00;                      // black
        } else {
          temp2 |= 0x04;                      // red
        }
        temp1 <<= 2;
        SendData(temp2);
      }
    }
    for (long k = 0; k < 160; k++) {                // 1/4 show white
      SendData(0x33);
    }

  }
  for (long i = 0; i < 192; i++) {
    for (long k = 0; k < 160; k++) {                 // 1/4 show red
      SendData(0x44);
    }
    for (long k = 0; k < 160; k++) {                 // 1/4 show black
      SendData(0x00);
    }
  }
  SendCommand(0x92);
  SendCommand(DISPLAY_REFRESH);
  DelayMs(100);
  WaitUntilIdle();
}

/**
    @brief: After this command is transmitted, the chip would enter the
            deep-sleep mode to save power.
            The deep sleep mode would return to standby by hardware reset.
            The only one parameter is a check code, the command would be
            executed if check code = 0xA5.
            You can use EPD_Reset() to awaken
*/
void Epd::Sleep(void) {
  SendCommand(POWER_OFF);
  WaitUntilIdle();
  SendCommand(DEEP_SLEEP);
  SendData(0xa5);
}

/* END OF FILE */
